"""
This module contains the functions to scrape the AHBs for Pruefidentifikatoren.
"""

import fnmatch
import gc
import re
from datetime import date
from pathlib import Path
from typing import Optional

import click
import docx
import tomlkit
from docx.table import Table
from efoli import EdifactFormatVersion

from kohlrahbi.ahbtable.ahbtable import AhbTable
from kohlrahbi.docxfilefinder import DocxFileFinder
from kohlrahbi.enums.ahbexportfileformat import AhbExportFileFormat
from kohlrahbi.logger import logger
from kohlrahbi.read_functions import (
    get_ahb_table,
    get_all_paragraphs_and_tables,
    table_header_starts_with_text_edifact_struktur,
)
from kohlrahbi.seed import Seed
from kohlrahbi.unfoldedahb import UnfoldedAhb
from kohlrahbi.unfoldedahb.unfoldedahbtable import are_equal_except_for_guids

_pruefi_pattern = re.compile(r"^[1-9]\d{4}$")


def load_pruefi_docx_file_map_from_file(path_to_pruefi_docx_file_map_file: Path) -> dict[str, dict[str, str]]:
    """
    Loads the file which contains all known Prüfidentifikatoren.
    The file may be manually updated with the script `collect_pruefis.py`.
    """
    assert path_to_pruefi_docx_file_map_file.exists(), f"The file {path_to_pruefi_docx_file_map_file} does not exist."

    with open(path_to_pruefi_docx_file_map_file, "rb") as file:
        pruefi_docx_file_map: dict[str, dict[str, str]] = tomlkit.load(file)

    return pruefi_docx_file_map


def process_ahb_table(
    ahb_table: AhbTable,
    pruefi: str,
    output_path: Path,
    file_type: tuple[AhbExportFileFormat, ...],
) -> None:
    """
    Process the ahb table.
    """
    unfolded_ahb = UnfoldedAhb.from_ahb_table(ahb_table=ahb_table, pruefi=pruefi)

    try:
        json_file_path = unfolded_ahb.get_flatahb_json_file_path(output_path)
        excel_file_path = unfolded_ahb.get_xlsx_file_path(output_path)
        csv_file_path = unfolded_ahb.get_csv_file_path(output_path)
    except ValueError:
        logger.warning("Error while determining file paths for pruefi '%s'. Skipping saving files.", pruefi)
        return
    pruefi_did_change_since_last_scraping: bool = True  # we assume it yes, if we can't compare or unless we know better
    if AhbExportFileFormat.FLATAHB in file_type and json_file_path.exists():
        # the flat ahb is the only file format from which we can READ to compare our current with previous results
        pruefi_did_change_since_last_scraping = not are_equal_except_for_guids(unfolded_ahb, json_file_path)
        logger.info("Pruefi '%s' did change since last scraping: %s", pruefi, pruefi_did_change_since_last_scraping)
    # ⚠ here we assume that the csv/json/xlsx files are in sync, if they exist.
    # this means: if the json file didn't change and a csv file exists, we expect the csv file to also be unchanged
    excel_needs_to_be_dumped = AhbExportFileFormat.XLSX in file_type and (
        (not excel_file_path.exists()) or pruefi_did_change_since_last_scraping
    )
    json_needs_to_be_dumped = AhbExportFileFormat.FLATAHB in file_type and (
        (not json_file_path.exists()) or pruefi_did_change_since_last_scraping
    )
    csv_needs_to_be_dumped = AhbExportFileFormat.CSV in file_type and (
        (not csv_file_path.exists()) or pruefi_did_change_since_last_scraping
    )
    if excel_needs_to_be_dumped:
        unfolded_ahb.dump_xlsx(output_path)
    if json_needs_to_be_dumped:
        unfolded_ahb.dump_flatahb_json(output_path)
    if csv_needs_to_be_dumped:
        unfolded_ahb.dump_csv(output_path)
    del unfolded_ahb


# pylint:disable=anomalous-backslash-in-string
def get_valid_pruefis(list_of_pruefis: list[str], all_known_pruefis: Optional[list[str]] = None) -> list[str]:
    """
    This function returns a list with only those pruefis which match the pruefi_pattern "^[1-9]\\d{4}$".
    It also supports unix wildcards like '*' and '?' if a list of known pruefis is given.
    E.g. '11*' for all pruefis starting with '11' or '*01' for all pruefis ending with '01'.
    """
    result: set[str] = set()

    for pruefi in list_of_pruefis:
        if ("*" in pruefi or "?" in pruefi) and all_known_pruefis:
            filtered_pruefis = fnmatch.filter(all_known_pruefis, pruefi)
            result = result.union(filtered_pruefis)
        elif _pruefi_pattern.match(pruefi):
            result.add(pruefi)

    return sorted(list(result))


def validate_pruefis(pruefis: list[str]) -> list[str]:
    """
    Validate the pruefi_to_file_mapping parameter.
    """
    valid_pruefis = get_valid_pruefis(pruefis)
    if not valid_pruefis:
        click.secho("⚠️ There are no valid pruefidentifkatoren.", fg="red")
        raise click.Abort()
    return valid_pruefis


def process_pruefi(
    pruefi: str,
    path_to_ahb_docx_file: Path,
    output_path: Path,
    file_type: tuple[AhbExportFileFormat, ...],
) -> None:
    """
    Process one pruefi.
    If the input path ends with .docx, we assume that the file containing the pruefi is given.
    Therefore, we only access that file.
    """

    # TODO try to cache document objects cause it is slow to read them from disk # pylint: disable=fixme
    doc = docx.Document(str(path_to_ahb_docx_file))

    if not doc:
        return

    ahb_table = get_ahb_table(document=doc, pruefi=pruefi)
    if not ahb_table:
        return

    process_ahb_table(ahb_table, pruefi, output_path, file_type)
    del ahb_table.table
    del ahb_table
    del doc
    gc.collect()


def get_ahb_documents_path(base_path: Path, version: str) -> Path:
    """Returns the path to the AHB documents for the specified format version."""
    path = base_path / f"edi_energy_de/{version}"
    if not path.exists():
        raise FileNotFoundError(f"The specified path {path.absolute()} does not exist.")
    return path


def find_pruefidentifikatoren(path: Path) -> dict[str, str]:
    """finds pruefis in given dir"""
    pruefis = {}

    ahb_file_finder = DocxFileFinder.from_input_path(input_path=path)
    ahb_file_finder.filter_for_latest_ahb_docx_files()

    for docx_path in ahb_file_finder.paths_to_docx_files:
        pruefis.update(extract_pruefis_from_docx(docx_path))
    return dict(sorted(pruefis.items()))


def extract_pruefis_from_docx(docx_path: Path) -> dict[str, str]:
    """Extracts the Prüfidentifikatoren from the given docx file."""
    doc = docx.Document(str(docx_path))
    pruefis: dict[str, str] = {}
    for item in get_all_paragraphs_and_tables(doc):
        if (
            isinstance(item, Table)
            and table_header_starts_with_text_edifact_struktur(item)
            and table_header_contains_text_pruefidentifikator(item)
        ):
            # pylint:disable=not-an-iterable
            pruefis.update({pruefi: docx_path.name for pruefi in extract_pruefis_from_table(item)})
    return pruefis


def save_pruefi_map_to_toml(pruefis: dict[str, str], version: str) -> None:
    """Saves the pruefis to file mapping to a toml file."""
    output_filename = f"{version}_pruefi_docx_filename_map.toml"
    output_file_path = Path(__file__).parents[1] / "cache" / output_filename
    output_file_path.parent.mkdir(parents=True, exist_ok=True)

    toml_data = {"meta_data": {"updated_on": date.today()}, "pruefidentifikatoren": pruefis}
    with open(output_file_path, "w", encoding="utf-8") as f:
        tomlkit.dump(toml_data, f)
        logger.info("🎉 Successfully updated %s and saved it at %s.", output_filename, output_file_path)


def log_no_pruefis_warning(version: str, path: Path) -> None:
    """Logs a warning if no Prüfidentifikatoren were found."""
    logger.warning("No Prüfidentifikatoren found in the AHB documents for format version %s at %s.", version, path)


def get_default_pruefi_map(path: Path) -> dict[str, str]:
    """Returns a default pruefi map if no pruefis were found."""
    return {"⚠️ No Prüfidentifikatoren found": f"No AHB documents found in {path}."}


def extract_pruefis_from_table(table: Table) -> list[str]:
    """Extracts the Prüfidentifikatoren from given table."""
    seed = Seed.from_table(docx_table=table)
    logger.info("Found a table with the following pruefis: %s", seed.pruefidentifikatoren)
    return seed.pruefidentifikatoren


def table_header_contains_text_pruefidentifikator(table: Table) -> bool:
    """Checks if the table header contains the text 'Prüfidentifikator'."""
    pattern = r"Prüfidentifikator(?:\t){0,10}\t\d+"
    # "matches "Prüfidentifikator" followed by at least 1 tab separated numbers, max 11 pruefis is chosen arbitrarily
    return bool(re.search(pattern, table.row_cells(0)[-1].text))


def get_pruefi_to_file_mapping(basic_input_path: Path, format_version: EdifactFormatVersion) -> dict[str, str]:
    """Returns the pruefi to file mapping. If the cache file does not exist, it creates it."""
    default_path_to_cache_file = Path(__file__).parents[1] / "cache" / f"{format_version}_pruefi_docx_filename_map.toml"

    if default_path_to_cache_file.exists():
        pruefi_to_file_mapping_cache = load_pruefi_docx_file_map_from_file(default_path_to_cache_file)
        pruefi_to_file_mapping = pruefi_to_file_mapping_cache.get("pruefidentifikatoren")
        if isinstance(pruefi_to_file_mapping, type(None)):
            raise ReferenceError(f"Could not find pruefidentifikatoren in {default_path_to_cache_file}")
        return dict(pruefi_to_file_mapping)

    path_to_docx_files = basic_input_path / Path(f"edi_energy_de/{format_version}")
    pruefi_to_file_mapping = find_pruefidentifikatoren(path_to_docx_files)
    save_pruefi_map_to_toml(pruefi_to_file_mapping, format_version.value)
    return pruefi_to_file_mapping


def reduce_pruefi_to_file_mapping(pruefi_to_file_mapping: dict[str, str], pruefis: list[str]) -> dict[str, str]:
    """
    If the user provided pruefis, we filter the pruefi_to_file_mapping for these pruefis.
    """
    return {pruefi: filename for pruefi, filename in pruefi_to_file_mapping.items() if pruefi in pruefis}


def remove_vanished_pruefis(pruefi_to_file_mapping: dict[str, str], output_path: Path) -> None:
    """
    Looks for existing pruefis in output_path and removes them if there are no present in pruefi_to_file_mapping
    """
    try:
        existing_pruefis = list_files_in_subdirs(output_path)
        for pruefi, file_path in existing_pruefis:
            if pruefi not in pruefi_to_file_mapping:
                logger.info("Removing pruefi '%s' from '%s' .", pruefi, file_path.parent)
                file_path.unlink()
            if is_parent_dir_empty(file_path):
                file_path.parent.rmdir()
                logger.info("Removing empty '%s' .", file_path.parent)
    except ValueError:
        logger.warning(
            "Error while cleaning existing files from given output_path '%s'. Skipping saving files.", output_path
        )


def is_parent_dir_empty(file_path: Path) -> bool:
    """
    Checks if a parent dir of a provided file path is empty
    """
    parent_dir = file_path.parent
    return not any(parent_dir.iterdir())


def list_files_in_subdirs(path: Path) -> list[tuple[str, Path]]:
    """
    Find all files in a given subdir.
    """
    files_list = []
    for file_path in path.rglob("*"):
        if file_path.is_file():
            file_name_without_ext = file_path.stem
            if file_name_without_ext not in ["conditions", "packages"]:
                files_list.append((file_name_without_ext, file_path))
    return files_list


# pylint: disable=too-many-arguments, too-many-positional-arguments
def scrape_pruefis(
    pruefis: list[str],
    basic_input_path: Path,
    output_path: Path,
    file_type: tuple[AhbExportFileFormat, ...],
    format_version: EdifactFormatVersion,
    clear_output_path: bool,
) -> None:
    """
    starts the scraping process for provided pruefi_to_file_mappings
    """
    pruefi_to_file_mapping = get_pruefi_to_file_mapping(
        basic_input_path=basic_input_path, format_version=format_version
    )
    if len(pruefis) > 0:
        validated_pruefis = validate_pruefis(pruefis)
        pruefi_to_file_mapping = reduce_pruefi_to_file_mapping(pruefi_to_file_mapping, validated_pruefis)
    if clear_output_path:
        remove_vanished_pruefis(pruefi_to_file_mapping, output_path)
    for pruefi, filename in pruefi_to_file_mapping.items():
        try:
            logger.info("start looking for pruefi '%s'", pruefi)
            if not filename:
                logger.warning("No filename for pruefi '%s' provided", pruefi)
                continue
            path_to_ahb_docx_file = (
                basic_input_path / Path("edi_energy_de") / Path(format_version.name) / Path(filename)
            )
            process_pruefi(pruefi, path_to_ahb_docx_file, output_path, file_type)
        except FileNotFoundError:
            logger.exception("File not found for pruefi '%s'", pruefi)
        # sorry for the pokemon catch
        except Exception as e:  # pylint: disable=broad-except
            logger.exception("Error processing pruefi '%s': %s", pruefi, str(e))



"""
This module defines an enum class representing the available export file formats for AhbExport.
"""

from enum import StrEnum


class AhbExportFileFormat(StrEnum):
    """
    Enum class representing the available export file formats for AhbExport.

    Attributes:
        FLATAHB (str): The FLATAHB file format.
        CSV (str): The CSV file format.
        XLSX (str): The XLSX file format.
    """

    FLATAHB = "flatahb"
    CSV = "csv"
    XLSX = "xlsx"



"""
This module contains the class EdifactStrukturCell
"""

import re

import pandas as pd
from docx.table import _Cell
from pydantic import BaseModel, ConfigDict

_segment_group_pattern = re.compile(r"^SG\d+$")
_segment_pattern = re.compile(r"^[A-Z]{3}$")
_data_element_pattern = re.compile(r"^\d{4}$")
_segment_id_pattern = re.compile(r"^[A-Z\d]\d{4}$")


# pylint: disable=too-few-public-methods
class EdifactStrukturCell(BaseModel):
    """
    EdifactStrukturCell contains all information and a method
    to extract the segment name, segment group, segment and data element.
    """

    table_cell: _Cell
    edifact_struktur_cell_left_indent_position: int

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def parse(self, ahb_row_dataframe: pd.DataFrame) -> pd.DataFrame:
        """Parses a paragraph in the edifact struktur column and puts the information into the appropriate columns

        Args:
            table_cell (Cell): edifact struktur cell
            dataframe (pd.DataFrame): Contains all infos
            row_index (int): Current index of the DataFrame
            edifact_struktur_cell_left_indent_position (int): Position of the left indent from the indicator edifact
                struktur cell
        """

        joined_text = " ".join(p.text for p in self.table_cell.paragraphs)
        splitted_text_at_tabs = joined_text.split("\t")

        row_index = ahb_row_dataframe.index.max()

        for text in splitted_text_at_tabs:
            if _segment_group_pattern.match(text):
                ahb_row_dataframe.at[row_index, "Segment Gruppe"] = text
            elif _segment_pattern.match(text):
                ahb_row_dataframe.at[row_index, "Segment"] = text
            elif _data_element_pattern.match(text):
                ahb_row_dataframe.at[row_index, "Datenelement"] = text
            elif _segment_id_pattern.match(text):
                ahb_row_dataframe.at[row_index, "Segment ID"] = text
            elif text != "":
                ahb_row_dataframe.at[row_index, "Segment Gruppe"] = text
        return ahb_row_dataframe



"""
This module collects all classes regarding to unfolded ahb table.
"""

from .unfoldedahbline import UnfoldedAhbLine
from .unfoldedahbtable import UnfoldedAhb
from .unfoldedahbtablemetadata import UnfoldedAhbTableMetaData

__all__ = ["UnfoldedAhb", "UnfoldedAhbLine", "UnfoldedAhbTableMetaData"]



"""
This __init__ module collects all cell classes
"""

from .bedinungscell import BedingungCell
from .bodycell import BodyCell
from .edifactstrukturcell import EdifactStrukturCell

__all__ = ["BedingungCell", "BodyCell", "EdifactStrukturCell"]



"""
This module contains the TableHeader class.
"""

from enum import StrEnum
from typing import Mapping, cast

from docx.table import _Cell
from docx.text.paragraph import Paragraph
from more_itertools import first, last
from pydantic import BaseModel, Field


class HeaderSection(StrEnum):
    """
    Enum for the different sections of the table header
    """

    BESCHREIBUNG = "beschreibung"
    KOMMUNIKATION_VON = "kommunikation von"
    PRUEFIDENTIFIKATOR = "pruefidentifikator"


def get_tabstop_positions(paragraph: Paragraph) -> list[int]:
    """Find all tabstop positions in a given paragraph.

    Mainly the tabstop positions of cells from the middle column are determined

    Args:
        paragraph (Paragraph):

    Returns:
        list[int]: All tabstop positions in the given paragraph
    """
    tabstop_positions: list[int] = []
    for tabstop in paragraph.paragraph_format.tab_stops:
        tabstop_positions.append(tabstop.position)
    return tabstop_positions


def create_mapping_of_tabstop_positions(
    initial_tabstop_positions: list[int],
    current_tabstop_positions: list[int],
) -> Mapping[int, int]:
    """
    Create a mapping of the tabstop positions of the Prüfidentifikatoren columns.

    For the current tabstop positions, the initial ones with the least difference are
    found to account for changes in tabstop positions between paragraphs.

    Returns:
        dict[int, int]: All initial tabstop positions mapped to the current tabstop positions
    """
    # create a mapping of the tabstop positions
    mapping: dict[int, int] = {}
    # Sort the lists in ascending order
    initial_tabstop_positions.sort()
    current_tabstop_positions.sort()

    # Iterate over the sorted lists and find the entries with the least difference
    for i, current_pos in enumerate(current_tabstop_positions):
        min_diff = float("inf")
        min_j = None
        for j, initial_pos in enumerate(initial_tabstop_positions):
            diff = abs(current_pos - initial_pos)
            if diff < min_diff:
                min_diff = diff
                min_j = j
        if min_j is None:
            raise ValueError("min_j should not be None")

        mapping[current_tabstop_positions[i]] = initial_tabstop_positions[min_j]

    return mapping


# pylint: disable=too-few-public-methods


class PruefiMetaData(BaseModel):
    """
    This class contains the information about the Prüfidentifikatoren
    """

    pruefidentifikator: str
    name: str
    communication_direction: str


class TableHeader(BaseModel):
    """
    Class for the table header.
    It contains the information about the Prüfidentifikatoren.
    """

    pruefi_meta_data: list[PruefiMetaData] = Field(default_factory=list)

    @classmethod
    def from_header_cell(cls, row_cell: _Cell) -> "TableHeader":
        """
        Create a TableHeader instance from a list of strings.
        """

        if not row_cell.paragraphs[-1].text.startswith("Prüfidentifikator"):
            raise ValueError("The last paragraph should start with 'Prüfidentifikator'")

        collector = cls.initialize_collector(paragraph=last(row_cell.paragraphs))

        section_type: HeaderSection

        # the main goal of this loop is to find the tabstop positions of each Prüfidentifikator.
        for paragraph in row_cell.paragraphs:
            splitted_text = paragraph.text.split("\t")
            text_prefix = first(splitted_text)

            # the paragraph with the Prüfidentifikatoren has slightly different tabstop positions compared to
            # the table content rows. So we skip this paragraph.
            does_paragraph_contain_pruefidentifikatoren = text_prefix == "Prüfidentifikator"
            if does_paragraph_contain_pruefidentifikatoren:
                continue

            # only the paragraphs with the Beschreibung and Kommunikation von have the same tabstop positions as
            # the table content rows.
            does_paragraph_contain_beschreibung_or_kommunikation_von = text_prefix in {
                "Beschreibung",
                "Kommunikation von",
            }
            if does_paragraph_contain_beschreibung_or_kommunikation_von:
                initial_tabstop_positions = get_tabstop_positions(paragraph=paragraph)
                tabstop_mapper = dict(zip(initial_tabstop_positions, collector.keys()))
                section_type = HeaderSection[text_prefix.replace(" ", "_").upper()]
                splitted_text.pop(0)
                for pruefi, text in zip(collector.keys(), splitted_text):
                    collector[pruefi][section_type.value] = text + " "
            else:
                if "" in splitted_text and len(splitted_text) > len(collector.keys()):
                    splitted_text.remove("")

                for tabstop_position, text in zip(initial_tabstop_positions, splitted_text):
                    pruefi = tabstop_mapper[tabstop_position]
                    collector[pruefi][section_type.value] += text + " "

        pruefi_meta_data = [
            PruefiMetaData(
                pruefidentifikator=pruefi,
                communication_direction=cls.ensure_single_space_between_words(
                    #  The cast function is a no-op at runtime and doesn't perform any actual type conversion;
                    #  it's only used for type checking purposes.
                    cast(str, meta_data[HeaderSection.KOMMUNIKATION_VON.value]),
                ),
                name=cls.ensure_single_space_between_words(cast(str, meta_data[HeaderSection.BESCHREIBUNG.value])),
            )
            for pruefi, meta_data in collector.items()
            if isinstance(meta_data[HeaderSection.KOMMUNIKATION_VON.value], str)
        ]

        return cls(pruefi_meta_data=pruefi_meta_data)

    @staticmethod
    def initialize_collector(paragraph: Paragraph) -> dict[str, dict[str, str | int]]:
        """Initialize the collector"""
        current_tabstop_positions = get_tabstop_positions(paragraph=paragraph)
        splitted_text = paragraph.text.split("\t")
        splitted_text.remove("Prüfidentifikator")

        collector: dict[str, dict[str, str | int]] = {
            pruefidentifikator: {
                HeaderSection.BESCHREIBUNG.value: "",
                HeaderSection.KOMMUNIKATION_VON.value: "",
                "tabstop_position": tab_stop,
            }
            for pruefidentifikator, tab_stop in zip(splitted_text, current_tabstop_positions)
        }

        if not collector:
            raise ValueError("collector should not be empty")

        return collector

    @staticmethod
    def ensure_single_space_between_words(text: str) -> str:
        """Inserts a single space between words"""
        return " ".join(text.split())

    def get_pruefidentifikatoren(self) -> list[str]:
        """
        Get all Prüfidentifikatoren from the table header.
        The order of the Prüfidentifikatoren is the same as in the docx table headers.
        So there should be no duplicates.
        """
        # pylint:disable=not-an-iterable
        return [pruefi.pruefidentifikator for pruefi in self.pruefi_meta_data]



# pylint:disable=too-few-public-methods
"""
This module contains classes that are required to model "Anwendungshandbücher" (AHB). There are two kinds of AHBs:
1. the "flat" AHB is very similar to the flat structure scraped from the official PDF files. It has no deep
structure.
2. the "nested" AHB which contains structural information (e.g. that a segment group is contained in
another segment group)
"""
import re
from typing import Annotated, Callable, Optional, Sequence, Union
from uuid import UUID

from more_itertools import last, split_when
from pydantic import BaseModel, Field, StringConstraints, field_validator

from kohlrahbi.models.edifact_components import (
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
)

_VERSION = "0.3.0"  #: version to be written into the deep ahb


# pylint:disable=too-many-instance-attributes
class AhbLine(BaseModel):
    """
    An AhbLine is a single line inside the machine-readable, flat AHB.
    """

    guid: Optional[UUID] = Field(default=None, description="optional key")
    # because the combination (segment group, segment, data element, name) is not guaranteed to be unique
    # yes, it's actually that bad already
    segment_group_key: Optional[str] = Field(default=None, description="the segment group, e.g. 'SG5'")

    segment_code: Optional[str] = Field(default=None, description="the segment, e.g. 'IDE'")

    data_element: Optional[str] = Field(default=None, description="the data element ID, e.g. '3224'")

    segment_id: Optional[str] = Field(
        default=None,
        description="the 5 digit segment id, e.g. '00003' for Nachrichten Kopfsegment. This is available since FV2410.",
    )

    value_pool_entry: Optional[str] = Field(
        None, description="one of (possible multiple) allowed values, e.g. 'E01' or '293'"
    )

    name: Optional[str] = Field(
        None, description="the name, e.g. 'Meldepunkt'. It can be both the description of a field but also its meaning"
    )

    # Check the unittest test_csv_file_reading_11042 to see the different values of name. It's not only the grey fields
    # where user input is expected but also the meanings / values of value pool entries. This means the exact meaning of
    # name can only be determined in the context in which it is used. This is one of many shortcoming of the current AHB
    # structure: Things in the same column don't necessarily mean the same thing.
    ahb_expression: Annotated[Optional[str], StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        default=None,
        description=(
            "a requirement indicator + an optional condition ('ahb expression'),"
            "e.g. 'Muss [123] O [456]'. Note: to parse expressions from AHBs"
            "consider using AHBicht: https://github.com/Hochfrequenz/ahbicht/"
        ),
    )
    conditions: Optional[str] = Field(
        default=None,
        description=(
            "The condition text describes the text to the optional condition of the ahb expression."
            "E.g. '[492] This is a condition text. [999] And this is another one.'"
        ),
    )

    section_name: Optional[str] = Field(
        default=None,
        description=(
            "The section name describes the purpose of a segment,"
            "e.g. 'Nachrichten-Kopfsegment' or 'Beginn der Nachricht'"
        ),
    )

    index: Optional[int] = Field(
        default=None,
        description=(
            "index is a number that describes the position of the AHBLine"
            "inside the original PDF- and FlatAnwendungshandbuch."
        ),
        ge=0,
    )

    def holds_any_information(self) -> bool:
        """
        Returns true iff the line holds any information exception for just a GUID.
        This is useful to filter out empty lines which are artefacts remaining from the scraping process.
        """
        # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
        # pylint: disable=no-member
        return (
            (self.segment_group_key is not None and len(self.segment_group_key.strip()) > 0)
            or (self.segment_code is not None and len(self.segment_code.strip()) > 0)
            or (self.data_element is not None and len(self.data_element.strip()) > 0)
            or (self.value_pool_entry is not None and len(self.value_pool_entry.strip()) > 0)
            or (self.name is not None and len(self.name.strip()) > 0)
            or (self.ahb_expression is not None and len(self.ahb_expression.strip()) > 0)
        )

    def get_discriminator(self, include_name: bool = True) -> str:
        """
        Generate a unique yet readable discriminator for this given line.
        This discriminator is generated just from the line itself without any information on where it occurs.
        It does neither know its position inside the AHB nor parent or sub groups.
        """
        result: str
        if self.segment_group_key:
            result = f"{self.segment_group_key}->"
        else:
            result = ""
        result += f"{self.segment_code}->{self.data_element}"
        if self.name and include_name:
            result += f" ({self.name})"
        return result


class AhbMetaInformation(BaseModel):
    """
    Meta information about an AHB like e.g. its title, Prüfidentifikator, possible sender and receiver roles
    """

    pruefidentifikator: str = Field(
        ..., description="identifies the message type (within a fixed format version) e.g. '11042' or '13012'"
    )
    maus_version: Optional[str] = Field(
        default=_VERSION, description="semantic version of maus used to create this document"
    )
    description: Annotated[Optional[str], StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        default=None,
        description="an optional description of the purpose of the pruefidentifikator; e.g. 'Anmeldung MSB' for 11042",
    )
    direction: Annotated[Optional[str], StringConstraints(strip_whitespace=True, min_length=1)] = Field(
        default=None,
        description=(
            "a stringly typed description of the roles of sender and receiver of the message"
            "(the row name in the AHB is 'Kommunikation von'); e.g. 'MSB an NB' for 11042"
        ),
    )


def _remove_grouped_ahb_lines_containing_section_name(
    grouped_ahb_lines: list[list[AhbLine]], section_name: str
) -> list[list[AhbLine]]:
    """
    Removes all groups of ahb lines that contain a line with the given section name and returns a new list instance.
    """
    return [goal for goal in grouped_ahb_lines if any(ahbl.section_name == section_name for ahbl in goal)]


# pylint:disable=unused-argument
def _check_that_nearly_all_lines_have_a_segment_group(  # type:ignore[no-untyped-def]
    value: list[AhbLine],
):
    """
    Loops over all provided ahb lines and checks that only at the beginning and the end there are lines without a
    segment group. In between every line has to have a segment group specified. This is necessary for the navigation
    to work because it primarily focuses and relies on correct SG information in the lines .
    """
    switches_from_no_sg_to_sg = list(
        split_when(value, lambda x, y: x.segment_group_key is None and y.segment_group_key is not None)
    )
    switches_from_no_sg_to_sg = _remove_grouped_ahb_lines_containing_section_name(
        grouped_ahb_lines=switches_from_no_sg_to_sg, section_name="Abschnitts-Kontrollsegment"
    )
    if len(switches_from_no_sg_to_sg) > 2:
        raise ValueError(f"There is a None segment group in line {last(switches_from_no_sg_to_sg[1])}")


_data_element_pattern = re.compile(r"^\d{4}$|^\d{5}$|^[A-Za-z]+\d{4}$")


# pylint:disable=unused-argument
def _check_that_line_has_either_none_or_d4_data_element(  # type:ignore[no-untyped-def]
    value: AhbLine,
):
    """
    checks that the given line has either a None data element or a data element that matches
    \\d{4},\\ d{5} or [A-Za-z]+\\d{4}.
    AhbLine(data_element = 0001), AhbLine(data_element = 00001) or AhbLine(data_element = R0001)
    """
    if value.data_element is None:
        return
    match = _data_element_pattern.match(value.data_element)
    if match is None:
        raise ValueError(f"The data_element '{value.data_element}' does not match {_data_element_pattern}")


_segment_group_key_pattern = re.compile(r"^SG\d+$")


# pylint:disable=unused-argument
def _check_that_line_has_either_none_or_matching_sg(value: AhbLine):  # type:ignore[no-untyped-def]
    """
    checks that the given line has a segment group key that is either None (for root) or matches SG\\d+
    """
    if value.segment_group_key is None:
        return
    match = _segment_group_key_pattern.match(value.segment_group_key)
    if match is None:
        raise ValueError(
            f"The segment_group_key '{value.segment_group_key}' does not match {_segment_group_key_pattern}"
        )


_segment_code_pattern = re.compile(r"^[A-Z]+$")


# pylint:disable=unused-argument
def _check_that_line_has_either_none_az_segment_code(  # type:ignore[no-untyped-def]
    value: AhbLine,
):
    """
    checks that the given line has either a None segment code or a segment code that consists of all upper letters
    """
    if value.segment_code is None:
        return
    match = _segment_code_pattern.match(value.segment_code)
    if match is None:
        raise ValueError(f"The segment_code '{value.segment_code}' does not match {_segment_code_pattern}")


class FlatAnwendungshandbuch(BaseModel):
    """
    A flat Anwendungshandbuch (AHB) models an Anwendungshandbuch as combination of some meta data and an ordered list
    of `.class:`.FlatAhbLine s. Basically a flat Anwendungshandbuch is the result of a simple scraping approach.
    You can create instances of this class without knowing anything about the "under the hood" structure of AHBs or MIGs
    """

    meta: AhbMetaInformation = Field(..., description="information about this AHB")

    lines: list[AhbLine] = Field(..., description="ordered list lines as they occur in the AHB")

    @field_validator("lines")
    @classmethod
    def validate_lines(cls, value: list[AhbLine]) -> list[AhbLine]:
        """
        The following checks are not baked into the AhbLine class itself, because they might be initialized
        with raw data that do not yet obey these strict validations. But as soon as we bundle them in a
        FlatAnwendungshandbuch, the sanitization shall be applied.
        """
        if value:
            _check_that_nearly_all_lines_have_a_segment_group(value)

        for ahb_line in value:
            _check_that_line_has_either_none_or_d4_data_element(ahb_line)
            _check_that_line_has_either_none_or_matching_sg(ahb_line)
            _check_that_line_has_either_none_az_segment_code(ahb_line)
        return value

    def get_segment_groups(self) -> list[Optional[str]]:
        """
        :return: a set with all segment groups in this AHB in the order in which they occur
        """
        return FlatAnwendungshandbuch._get_available_segment_groups(self.lines)

    @staticmethod
    def _get_available_segment_groups(lines: list[AhbLine]) -> list[Optional[str]]:
        """
        extracts the distinct segment groups from a list of ahb lines
        :param lines:
        :return: distinct segment groups, including None in the order in which they occur
        """
        result: list[Optional[str]] = []
        for line in lines:
            if line.segment_group_key not in result:
                # an "in" check against a set would be faster but we want to preserve both order and readability
                result.append(line.segment_group_key)
        return result

    def sort_lines_by_segment_groups(self):  # type:ignore[no-untyped-def]
        """
        sorts lines by segment groups while preserving the order inside the groups and the order between the groups.
        """
        self.lines = FlatAnwendungshandbuch._sorted_lines_by_segment_groups(self.lines, self.get_segment_groups())

    @staticmethod
    def _sorted_lines_by_segment_groups(ahb_lines: Sequence[AhbLine], sg_order: list[Optional[str]]) -> list[AhbLine]:
        """
        Calls sorted(...) on the provided list and returns a new list.
        Its purpose is, that if a segment group in the AHB (read from top to bottom in the flat ahb/pdf) is interrupted
        by other segment groups, the lines belonging to the same group will be next to each other.
        This is useful to later use a groupby aggregation that only returns one group key per segment group.

        The sort is stable such that the existing order inside the segment groups is maintained.

        Note that this also means, that the order of the return lines no longer the same as in the flat AHB.

        If you provide this function a list:
        [
            (SG2, Foo),
            (SG2, Bar),
            (SG3, ABC),
            (SG3, DEF),
            (SG2, SomethingElse)
        ]
        The result will be:
        [
            (SG2, Foo),
            (SG2, Bar),
            (SG2, SomethingElse)
            (SG3, ABC),
            (SG3, DEF)
        ]

        See the unittests for details.
        """

        # this code is in a static method to make it easily accessible for fine-grained unit testing
        result: list[AhbLine] = sorted(ahb_lines, key=lambda x: x.segment_group_key or "")
        result.sort(key=lambda ahb_line: sg_order.index(ahb_line.segment_group_key))
        return result


class DeepAhbInputReplacement(BaseModel):
    """
    A container class that models replacements of inputs in the DeepAnwendungshandbuch
    """

    #: true iff a replacement is applicable
    replacement_found: bool = Field(..., description="True iff a replacement is applicable")

    # replacements for DataElementValuePool
    value_pool_replacement: Optional[ValuePoolEntry] = Field(
        default=None,
        description=(
            "The replacement for a value pool entry."
            "Note that the replacements may be None even if replacements are found."
        ),
    )
    # replacements for DataElementFreeText
    free_text_replacement: Optional[DataElementFreeText] = Field(
        default=None,
        description=(
            "The replacement for a DataElementFreeText."
            "Note that the replacements may be None even if replacements are found."
        ),
    )


class DeepAnwendungshandbuch(BaseModel):
    """
    The data of the AHB nested as described in the MIG.
    """

    meta: AhbMetaInformation = Field(..., description="Information about this AHB")

    lines: list[SegmentGroup] = Field(..., description="The nested data")

    def reset_ahb_line_index(self) -> None:
        """
        reset the ahb line index for all lines in the DeepAnwendungshandbuch
        :return: nothing; modifies the instance
        """
        for line in self.lines:
            line.reset_ahb_line_index()

    @staticmethod
    def _query_segment_group(
        segment_group: SegmentGroup, predicate: Callable[[SegmentGroup], bool]
    ) -> list[SegmentGroup]:
        """
        recursively search for a segment group that matches the predicate
        :return: return empty list if nothing was found, the matching segment groups otherwise
        """
        result: list[SegmentGroup] = []
        if predicate(segment_group):
            result.append(segment_group)
        if segment_group.segment_groups is not None:
            for sub_group in segment_group.segment_groups:
                sub_result = DeepAnwendungshandbuch._query_segment_group(sub_group, predicate)
                result += sub_result
        return result

    def find_segment_groups(self, predicate: Callable[[SegmentGroup], bool]) -> list[SegmentGroup]:
        """
        recursively search for segment group in this ahb that meets the predicate.
        :return: list of segment groups that match the predicate; empty list otherwise
        """
        result: list[SegmentGroup] = []
        for line in self.lines:
            if line.segment_groups is not None:
                for segment_group in line.segment_groups:
                    result += DeepAnwendungshandbuch._query_segment_group(segment_group, predicate)
            for line_result in DeepAnwendungshandbuch._query_segment_group(line, predicate):
                if line_result not in result:
                    result.append(line_result)
        return result

    def find_segments(
        self,
        group_predicate: Callable[[SegmentGroup], bool] = lambda _: True,
        segment_predicate: Callable[[Segment], bool] = lambda _: True,
    ) -> list[Segment]:
        """
        recursively search for segment characterized by the segment_predicate inside a group characterized by the
        group_predicate.
        :return: list of matching segments, empty list if nothing was found
        """
        result: list[Segment] = []
        for segment_group in self.find_segment_groups(group_predicate):
            result += segment_group.find_segments(segment_predicate)
        for line in self.lines:
            if line.segments is not None:
                for segment in line.segments:
                    if segment_predicate(segment):
                        result.append(segment)
        return result

    def replace_inputs_based_on_discriminator(self, replacement_func: Callable[[str], DeepAhbInputReplacement]) -> None:
        """
        Replace all the entered_inputs in the entire DeepAnwendungshandbuch using the given replacement_func.
        Note that this modifies this DeepAnwendungshandbuch instance (self).
        """
        _replace_inputs_based_on_discriminator(self.lines, replacement_func)

    def get_all_value_pools(self) -> list[DataElementValuePool]:
        """
        recursively find all value pools in the deep ahb
        :return: a list of all value pools
        """
        result: list[DataElementValuePool] = []
        added_discriminators: set[Optional[str]] = set()
        # checks like "str in set" are way faster than "value pool in list"

        def add_to_result(value_pool: DataElementValuePool):  # type:ignore[no-untyped-def]
            if value_pool.discriminator in added_discriminators:
                # assumption: the discriminator is truly unique in an AHB
                return
            if value_pool not in result:
                added_discriminators.add(value_pool.discriminator)
                result.append(value_pool)

        for segment in self.find_segments():
            for sub_result in segment.get_all_value_pools():
                add_to_result(sub_result)
        for segment_group in self.find_segment_groups(lambda _: True):
            for sub_result in segment_group.get_all_value_pools():
                add_to_result(sub_result)
        return list(result)

    def get_all_expressions(self) -> list[str]:
        """
        recursively iterate through the deep ahb and return all distinct expressions found
        """
        result: set[str] = set()
        for segment in self.find_segments():
            if segment.ahb_expression:
                result.add(segment.ahb_expression)
            for data_element in segment.data_elements:
                if isinstance(data_element, DataElementFreeText):
                    if data_element.ahb_expression:
                        result.add(data_element.ahb_expression)
                elif isinstance(data_element, DataElementValuePool):
                    for value_pool_entry in data_element.value_pool:
                        if value_pool_entry.ahb_expression:
                            result.add(value_pool_entry.ahb_expression)
        for segment_group in self.find_segment_groups(lambda _: True):
            if segment_group.ahb_expression:
                result.add(segment_group.ahb_expression)
        return sorted(result)


def _replace_inputs_based_on_discriminator(
    segment_groups: list[SegmentGroup], replacement_func: Callable[[str], DeepAhbInputReplacement]
) -> None:
    """
    Replace all the entered_inputs in the entire list of segment groups using the given replacement_func.
    """

    def process_data_element(
        data_element: Union[DataElementFreeText, DataElementValuePool], replacement_result: DeepAhbInputReplacement
    ) -> None:
        if isinstance(data_element, DataElementFreeText) and replacement_result.free_text_replacement:
            data_element.free_text = replacement_result.free_text_replacement.free_text
        elif isinstance(data_element, DataElementValuePool) and replacement_result.value_pool_replacement:
            data_element.value_pool.append(replacement_result.value_pool_replacement)

    def process_segment(segment: Segment) -> None:
        for data_element in segment.data_elements:
            if data_element.discriminator:
                replacement_result = replacement_func(data_element.discriminator)
                if replacement_result.replacement_found:
                    process_data_element(data_element, replacement_result)

    for segment_group in segment_groups:
        if segment_group.segment_groups:
            _replace_inputs_based_on_discriminator(segment_group.segment_groups, replacement_func)

        if segment_group.segments:
            for segment in segment_group.segments:
                process_segment(segment)



"""
class which contains AHB package condition table
"""

import json
import re
from pathlib import Path

import pandas as pd
from docx.table import Table as DocxTable
from efoli import EdifactFormat
from pydantic import BaseModel, ConfigDict, Field

from kohlrahbi.ahbtable.ahbcondtions import parse_conditions_from_string
from kohlrahbi.logger import logger


class AhbPackageTable(BaseModel):
    """
    This class contains the AHB Package table as you see it in the beginning AHB documents,
    but in a machine readable format.
    Caution: if two PackageTables objects are combined so far only the package_dict field is updated.
    """

    table: pd.DataFrame = pd.DataFrame()
    package_dict: dict[EdifactFormat, dict[str, str]] = Field(default_factory=dict)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_docx_table(cls, docx_tables: list[DocxTable]) -> "AhbPackageTable":
        """
        Create an AhbPackageTable object from a docx table.
        """
        table_data = []
        for table in docx_tables:
            for row in table.rows:
                row_data = [cell.text for cell in row.cells]
                table_data.append(row_data)

        headers = table_data[0]
        data = table_data[1:]
        df = pd.DataFrame(data, columns=headers)
        return cls(table=df)

    def provide_conditions(self, edifact_format: EdifactFormat) -> dict[EdifactFormat, dict[str, str]]:
        """collect conditions from package table and store them in conditions dict."""
        conditions_dict: dict[EdifactFormat, dict[str, str]] = {edifact_format: {}}
        there_are_conditions = (self.table["Bedingungen"] != "").any()
        if there_are_conditions:
            for conditions_text in self.table["Bedingungen"][self.table["Bedingungen"] != ""]:
                conditions_dict = parse_conditions_from_string(conditions_text, edifact_format, conditions_dict)
        logger.info("The package conditions for %s were collected.", edifact_format)
        return conditions_dict

    def provide_packages(self, edifact_format: EdifactFormat) -> None:
        """collect conditions from package table and store them in conditions dict."""
        package_dict: dict[EdifactFormat, dict[str, str]] = {edifact_format: {}}

        there_are_packages = (self.table["Paket"] != "").any()
        if there_are_packages:
            for _, row in self.table.iterrows():
                package = row["Paket"]
                # Use re.search to find the first match
                match = re.search(r"\[(\d+P)\]", package)
                if not match:
                    raise ValueError("No valid package key found in the package column.")
                    # Extract the matched digits
                package = match.group(1)
                if package != "1P":
                    package_conditions = row["Paketvoraussetzung(en)"].strip()
                    # check whether package was already collected:
                    existing_text = package_dict[edifact_format].get(package)
                    is_package_key_collected_yet = existing_text is not None
                    key_exits_but_shorter_text = existing_text is not None and len(package_conditions) > len(
                        existing_text
                    )
                    if not is_package_key_collected_yet or key_exits_but_shorter_text:
                        package_dict[edifact_format][package] = package_conditions

        logger.info("Packages for %s were collected.", edifact_format)
        self.package_dict = package_dict

    def include_package_dict(self, to_add: dict[EdifactFormat, dict[str, str]] | None) -> None:
        """Include a dict of conditions to the conditions_dict"""
        if to_add is None:
            logger.info("Packages dict to be added is empty.")
            return
        for edifact_format, edi_cond_dict in to_add.items():
            for package_key, package_conditions in edi_cond_dict.items():
                if edifact_format in self.package_dict:
                    if (
                        package_key in self.package_dict[edifact_format]
                        and len(package_conditions) > len(self.package_dict[edifact_format][package_key])
                        or package_key not in self.package_dict[edifact_format]
                    ):
                        self.package_dict[edifact_format][package_key] = package_conditions
                else:
                    self.package_dict[edifact_format] = {package_key: package_conditions}

        logger.info("Packages were updated.")

    def dump_as_json(self, output_directory_path: Path) -> None:
        """
        Writes all collected packages to a json file.
        The file will be stored in the directory:
            'output_directory_path/<edifact_format>/conditions.json'
        """
        package_suffix = "P"
        package_prefix = "UB"
        for edifact_format, format_pkg_dict in self.package_dict.items():
            package_json_output_directory_path = output_directory_path / str(edifact_format)
            package_json_output_directory_path.mkdir(parents=True, exist_ok=True)
            file_path = package_json_output_directory_path / "packages.json"
            # resort  PackageKeyConditionTextMappings for output
            package_dict_p = {k: v for k, v in format_pkg_dict.items() if k.endswith(package_suffix)}
            package_dict_ub = {k: v for k, v in format_pkg_dict.items() if k.startswith(package_prefix)}
            sorted_package_dict_p = {
                k: package_dict_p[k] for k in sorted(package_dict_p, key=lambda x: int(x[: -len(package_suffix)]))
            }
            sorted_package_dict_ub = {
                k: package_dict_ub[k] for k in sorted(package_dict_ub, key=lambda x: int(x[len(package_prefix) :]))
            }

            sorted_package_dict = {**sorted_package_dict_ub, **sorted_package_dict_p}
            array = [
                {"package_key": i, "package_expression": sorted_package_dict[i], "edifact_format": edifact_format}
                for i in sorted_package_dict
            ]
            with open(file_path, "w", encoding="utf-8") as file:
                json.dump(array, file, ensure_ascii=False, indent=2)

            logger.info(
                "The package.json file for %s is saved at %s",
                edifact_format,
                file_path,
            )



"""
This module contains the functions to scrape the AHBs for quality maps.
"""

from pathlib import Path
from typing import Optional

import docx
import pandas as pd
from docx.document import Document
from docx.table import Table

from kohlrahbi.docxfilefinder import DocxFileFinder
from kohlrahbi.logger import logger
from kohlrahbi.qualitymap.qualitymaptable import QualityMapTable
from kohlrahbi.read_functions import get_all_paragraphs_and_tables


def find_docx_files(input_path: Path) -> list[Path]:
    """
    Find all .docx files containing quality maps.
    """
    docx_file_finder = DocxFileFinder.from_input_path(input_path=input_path)
    return docx_file_finder.get_docx_files_which_contain_quality_map()


def is_quality_map_table(table: Table) -> bool:
    """
    Checks if the given table is quality map table.
    """
    try:
        return table.cell(row_idx=0, col_idx=0).text.strip() == "Qualität\n\nSegmentgruppe"
    except IndexError:
        return False


def get_quality_map_table(document: Document) -> Optional[QualityMapTable]:
    """
    Reads a docx file and extracts the quality map table.
    Returns None if no such table was found.
    """

    # Iterate through the whole word document
    logger.info("🔁 Start iterating through paragraphs and tables")
    for table in document.tables:
        if is_quality_map_table(table=table):
            change_history_table = QualityMapTable.from_docx_quality_map_table(docx_table=table)
            return change_history_table

    return None


def process_docx_file(file_path: Path) -> Optional[QualityMapTable]:
    """
    Read and process quality map from a .docx file.
    """
    doc = docx.Document(str(file_path))
    logger.info("🤓 Start reading docx file '%s'", str(file_path))
    quality_map_table = get_quality_map_table(document=doc)

    if quality_map_table is not None:
        return quality_map_table
    return None


def scrape_quality_map(input_path: Path, output_path: Path) -> None:
    """
    starts the scraping process of the quality maps.

    Args:
        input_path (Path): The path to the input **directory** containing the docx files with quality maps.
        output_path (Path): The path to the output **directory** to save the scraped quality maps
    """
    if input_path.is_dir() is False:
        logger.error("The input path '%s' is not a directory.", input_path)
        raise NotADirectoryError(f"The input path '{input_path}' is not a directory.")

    logger.info("👀 Start looking for quality maps")
    ahb_file_paths = find_docx_files(input_path)

    for file_path in ahb_file_paths:
        quality_map_table = process_docx_file(file_path)
        logger.info("Done with %s", file_path)
        if quality_map_table is None:
            logger.info("No quality map table found in %s", file_path)
            continue
        quality_map_table.save_to_csv(output_path / Path(f"{file_path.stem}_quality_map.csv"))
        quality_map_table.save_to_xlsx(output_path / Path(f"{file_path.stem}_quality_map.xlsx"))



"""
This module provides the ChangeHistoryTable class
"""

import pandas as pd
from docx.table import Table
from pydantic import BaseModel, ConfigDict

from kohlrahbi.ahbtable.ahbsubtable import AhbSubTable


class ChangeHistoryTable(BaseModel):
    """
    This class  contains the change history table.
    """

    table: pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_docx_change_history_table(cls, docx_table: Table) -> "ChangeHistoryTable":
        """
        Create a ChangeHistorySubTable object from a change history table.
        """

        change_history_rows: list[list[str]] = []

        for row in docx_table.rows:
            sanitized_cells = list(AhbSubTable.iter_visible_cells(row=row))

            is_header_row = sanitized_cells[0].text == "Änd-ID" or sanitized_cells[2].text == "Bisher"
            if is_header_row:
                continue
            change_history_rows.append([cell.text for cell in sanitized_cells])

        headers = ["Änd-ID", "Ort", "Änderungen Bisher", "Änderungen Neu", "Grund der Anpassung", "Status"]

        df = pd.DataFrame(change_history_rows, columns=headers)

        return cls(table=df)

    def sanitize_table(self) -> None:
        """
        Sanitizes the change history table.
        Thanks to the page breaks in the docx file there are rows which are just a small part of the upper row.
        This function merges these rows.
        """

        def is_empty(val: str) -> bool:
            """
            Checks if the given value is considered empty for our case.
            """
            return pd.isna(val) or val == ""

        # Define a function to check if a value is considered empty for our case
        def is_the_first_column_empty(row: pd.Series) -> bool:  # type:ignore[type-arg]
            """
            Checks if the first column of the given row is empty.
            This is our indicator if the current row is a continuation of the upper row.
            """
            return is_empty(row.iloc[0])

        # Iterate over the DataFrame rows in reverse (to avoid skipping rows after removing them)
        for i in reversed(range(1, len(self.table))):
            # Check if the first and second columns of the current row are empty
            if is_the_first_column_empty(self.table.iloc[i]):
                # Merge with the upper row by concatenating the non-empty values
                for col in range(len(self.table.columns)):
                    cell_entry = self.table.iloc[i, col]
                    if isinstance(cell_entry, str) and not is_empty(cell_entry):
                        self.table.iloc[i - 1, col] = (
                            str(self.table.iloc[i - 1, col]) + " " + str(self.table.iloc[i, col]).strip()
                        )
                # Drop the current row since it has been merged
                self.table.drop(i, inplace=True)

        # Reset index after dropping rows
        self.table.reset_index(drop=True, inplace=True)



"""
Command line interface for the changehistory command.
"""

from pathlib import Path

import click
from efoli import EdifactFormatVersion

from kohlrahbi.ahb.command import check_python_version, validate_path
from kohlrahbi.changehistory import scrape_change_histories


@click.command()
@click.option(
    "-eemp",
    "--edi-energy-mirror-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path),
    help="The root path to the edi_energy_mirror repository.",
    required=True,
)
@click.option(
    "-o",
    "--output-path",
    type=click.Path(exists=False, dir_okay=True, file_okay=False, resolve_path=True, path_type=Path),
    callback=validate_path,
    default="output",
    prompt="Output directory",
    help="Define the path where you want to save the generated files.",
)
@click.option(
    "--format-version",
    multiple=False,
    type=click.Choice([e.value for e in EdifactFormatVersion], case_sensitive=False),
    help="Format version(s) of the AHB documents, e.g. FV2310",
)
@click.option(
    "--assume-yes",
    "-y",
    is_flag=True,
    default=False,
    help="Confirm all prompts automatically.",
)
def changehistory(
    edi_energy_mirror_path: Path,
    output_path: Path,
    format_version: EdifactFormatVersion | str,
    assume_yes: bool,  # pylint: disable=unused-argument
    # it is used by the callback function of the output-path
) -> None:
    """
    Scrape change histories from the input path and save them to the output path.

    Args:
        edi_energy_mirror_path (Path): The path to the input file or directory containing change histories.
        output_path (Path): The path to save the scraped change histories.
        format_version (EdifactFormatVersion | str): The version of the EDIFACT format to use for scraping.
            Can be either an instance of EdifactFormatVersion or a string representation of the version.
        assume_yes (bool): Flag indicating whether to assume "yes" for all prompts.
    """
    check_python_version()
    if isinstance(format_version, str):
        format_version = EdifactFormatVersion(format_version)
    input_path = edi_energy_mirror_path / "edi_energy_de" / format_version.value
    scrape_change_histories(input_path=input_path, output_path=output_path)



"""
This module contains a command-line interface for handling conditions.

The `conditions` function is the main entry point for the command-line interface.
It takes several options as command-line arguments and performs the necessary operations based on those options.

Options:
- `--edi-energy-mirror-path`: Specifies the path to the directory with an edi_energy_mirror repository structure.
- `--output-path`: Specifies the path where the generated files will be saved.
- `--format-version`: Specifies the format version(s) of the AHB documents.
- `--assume-yes`: Automatically confirms all prompts.

Usage:
To use this command-line interface, run the script and provide the necessary options as command-line arguments.

Example:
$ python command.py --eemp /path/to/eemp --output-path /path/to/output --format-version FV2310 --assume-yes
"""

# pylint: disable=duplicate-code
# duplicate-code warning is disabled because the code cli structure of 'ahb' and 'conditions' are intentionally similar.
import sys
from pathlib import Path

import click
from efoli import EdifactFormatVersion

from kohlrahbi.conditions import scrape_conditions


def check_python_version() -> None:
    """
    Check if the Python interpreter is greater or equal to 3.11
    """
    if sys.version_info.major != 3 or sys.version_info.minor < 11:
        raise click.Abort(
            f"""Python >=3.11 is required to run this script but you use Python
{sys.version_info.major}.{sys.version_info.minor}"""
        )


# pylint: disable=unused-argument
def validate_path(ctx, param, value) -> Path:  # type:ignore[no-untyped-def]
    """
    Ensure the path exists or offer to create it.
    """
    path = Path(value)
    if not path.exists():
        if ctx.params.get("assume_yes") or click.confirm(
            f"The path {value} does not exist. Would you like to create it?"
        ):
            path.mkdir(parents=True)
            click.secho(f"Created directory {path}.", fg="green")
        else:
            click.secho("👋 Alright I will end this program now. Have a nice day.", fg="green")
            raise click.Abort()
    return path


@click.command()
@click.option(
    "-eemp",
    "--edi-energy-mirror-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path),
    help="The root path to the edi_energy_mirror repository.",
    required=True,
)
@click.option(
    "-o",
    "--output-path",
    type=click.Path(exists=False, dir_okay=True, file_okay=False, resolve_path=True, path_type=Path),
    callback=validate_path,
    default="output",
    prompt="Output directory",
    help="""Define the path where you want to save the generated files. If the path does not exist,
you will be asked if you want to create it.""",
)
@click.option(
    "--format-version",
    multiple=False,
    type=click.Choice([e.value for e in EdifactFormatVersion], case_sensitive=False),
    required=True,
    help="Format version(s) of the AHB documents, e.g. FV2310",
)
@click.option(
    "--assume-yes",
    "-y",
    is_flag=True,
    default=False,
    help="Confirm all prompts automatically.",
)
def conditions(
    edi_energy_mirror_path: Path, output_path: Path, format_version: EdifactFormatVersion | str, assume_yes: bool
) -> None:
    """
    Scrape AHB documents for conditions.
    """
    check_python_version()
    if isinstance(format_version, str):
        format_version = EdifactFormatVersion(format_version)

    scrape_conditions(
        basic_input_path=edi_energy_mirror_path,
        output_path=output_path,
        format_version=format_version,
    )



# pylint: disable=line-too-long
"""
Contains conditions for times in allgemeine Festlegungen.
"""
# We decided against adding the time_packages to the regular packages.
# The time-packages are resolved by a special transformer in AHBicht:
# https://github.com/Hochfrequenz/ahbicht/blob/c51c81d2be098dd79ff52b754979892396207fe2/src/ahbicht/expressions/expression_resolver.py#L149

time_conditions = {
    "490": "wenn Wert in diesem DE, an der Stelle CCYYMMDDHHMM ein Zeitpunkt aus dem angegeben Zeitraum der Tabelle Kapitel 3.5 „Übersicht gesetzliche deutsche Sommerzeit (MESZ)“ der Spalten:\n›\t„Sommerzeit (MESZ) von“ Darstellung in UTC und\n›\t„Sommerzeit (MESZ) bis“ Darstellung in UTC ist.",
    "491": "wenn Wert in diesem DE, an der Stelle CCYYMMDDHHMM ein Zeitpunkt aus dem angegeben Zeitraum der Tabelle Kapitel 3.6 „Übersicht gesetzliche deutsche Zeit (MEZ)“ der Spalten: \n›\t„Winterzeit (MEZ) von“ Darstellung in UTC und\n›\t„Winterzeit (MEZ) bis“ Darstellung in UTC ist.",
    "492": "wenn MP-ID in NAD+MR aus Sparte Strom",
    "493": "wenn MP-ID in NAD+MR aus Sparte Gas",
    "931": "Format: ZZZ = +00",
    "932": "Format: HHMM = 2200",
    "933": "Format: HHMM = 2300",
    "934": "Format: HHMM = 0400",
    "935": "Format: HHMM = 0500",
}
